/* Common operations.

Copyright (C) 2015, 2016, 2017, 2018, 2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __OPS_H__
#define __OPS_H__

#include "types.h"
#include <string.h> /* for __COPY */

/* Get object reference from attribute. */

__ref __VALUE(__attr attr);

/* Direct access and manipulation of static objects. */

__attr __load_static_ignore(__ref obj);
__attr __load_static_replace(__attr __context, __attr context, __ref obj);
__attr __load_static_test(__attr __context, __attr context, __ref obj);

/* Direct retrieval operations, returning attribute locations. */

__attr *__get_class_attr_ref__(__ref obj, int pos);

#define __get_object_attr_ref__(OBJ, POS) (&(OBJ)->attrs[POS])

#define __get_object_attr_ref(OBJ, ATTRNAME) (__get_object_attr_ref__(OBJ, __ATTRPOS(ATTRNAME)))
#define __get_class_attr_ref(OBJ, ATTRNAME) (__get_class_attr_ref__(OBJ, __ATTRPOS(ATTRNAME)))

/* Attribute location operations. */

#define __load_via_attr_ref(ATTR) (*(ATTR))
#define __store_via_attr_ref(ATTR, VALUE) __set_attr(ATTR, VALUE)

#define __store_via_attr_ref__(ATTR, VALUE) (*(ATTR) = VALUE)

/* Direct retrieval operations, returning attributes. */

__attr __load_via_class__(__ref obj, int pos);
__attr __load_via_object__(__ref obj, int pos);
__attr __get_class_and_load__(__ref obj, int pos);

#define __load_via_class(OBJ, ATTRNAME) (__load_via_class__(OBJ, __ATTRPOS(ATTRNAME)))
#define __load_via_object(OBJ, ATTRNAME) (__load_via_object__(OBJ, __ATTRPOS(ATTRNAME)))
#define __get_class_and_load(OBJ, ATTRNAME) (__get_class_and_load__(OBJ, __ATTRPOS(ATTRNAME)))

/* Introspection. */

int __is_instance(__ref obj);
int __is_subclass(__ref obj, __attr cls);
int __is_instance_subclass(__ref obj, __attr cls);
int __is_type_instance(__ref obj);
__ref __get_class(__ref obj);
__attr __get_class_attr(__ref obj);

/* Attribute testing operations. */

__ref __test_specific_instance(__ref obj, __ref type);
__ref __test_specific_object(__ref obj, __ref type);
__ref __test_specific_type(__ref obj, __ref type);

__ref __test_common_instance__(__ref obj, int pos, int code);
__ref __test_common_object__(__ref obj, int pos, int code);
__ref __test_common_type__(__ref obj, int pos, int code);

#define __to_error(REF) (REF ? REF : (__raise_type_error(), (__ref) 0))

#define __test_common_instance(OBJ, TYPENAME) (__test_common_instance__(OBJ, __ATTRPOS(TYPENAME), __ATTRCODE(TYPENAME)))
#define __test_common_object(OBJ, TYPENAME) (__test_common_object__(OBJ, __ATTRPOS(TYPENAME), __ATTRCODE(TYPENAME)))
#define __test_common_type(OBJ, TYPENAME) (__test_common_type__(OBJ, __ATTRPOS(TYPENAME), __ATTRCODE(TYPENAME)))

/* Attribute testing and location operations, returning the address of the
   attribute as opposed to its value. */

__attr *__check_and_get_object_attr_ref_null(__ref obj, int pos, int code);

__attr *__check_and_get_object_attr_ref__(__ref obj, int pos, int code);

#define __check_and_get_object_attr_ref(OBJ, ATTRNAME) (__check_and_get_object_attr_ref__(OBJ, __ATTRPOS(ATTRNAME), __ATTRCODE(ATTRNAME)))

/* Attribute testing and retrieval operations. */

__attr __check_and_load_via_object_null(__ref obj, int pos, int code);

__attr __check_and_load_via_class__(__ref obj, int pos, int code);
__attr __check_and_load_via_object__(__ref obj, int pos, int code);
__attr __check_and_load_via_any__(__ref obj, int pos, int code);

#define __check_and_load_via_class(OBJ, ATTRNAME) (__check_and_load_via_class__(OBJ, __ATTRPOS(ATTRNAME), __ATTRCODE(ATTRNAME)))
#define __check_and_load_via_object(OBJ, ATTRNAME) (__check_and_load_via_object__(OBJ, __ATTRPOS(ATTRNAME), __ATTRCODE(ATTRNAME)))
#define __check_and_load_via_any(OBJ, ATTRNAME) (__check_and_load_via_any__(OBJ, __ATTRPOS(ATTRNAME), __ATTRCODE(ATTRNAME)))

/* Context-related operations. */

int __test_context_update(__attr context, __attr attr, int invoke);
__attr __test_context_revert(int target, __attr context, __attr attr, __attr contexts[]);
__attr __test_context_static(int target, __attr context, __ref value, __attr contexts[]);

__attr __test_context(__attr __context, __attr context, __attr attr);
__attr __update_context(__attr __context, __attr context, __attr attr);

#define __get_accessor(__TARGET) (__tmp_values[__TARGET])
#define __get_attr_ref(__TARGET) (__tmp_attr_refs[__TARGET])
#define __get_context(__TARGET) (__tmp_contexts[__TARGET])
#define __set_context(__TARGET, __ATTR) (__tmp_contexts[__TARGET] = (__ATTR))
#define __set_private_context(__ATTR) (__tmp_private_context = (__ATTR))
#define __set_accessor(__TARGET, __ATTR) (__tmp_values[__TARGET] = (__ATTR))
#define __set_target_accessor(__ATTR) (__tmp_target_value = (__ATTR))
#define __set_attr_ref(__TARGET, __ATTR) (__tmp_attr_refs[__TARGET] = (__ATTR))

/* Context testing for invocations. */

__attr __unwrap_callable(__attr callable);
__attr (*__get_function_unchecked(__attr target))();
__attr (*__get_function(__attr context, __attr target))();
__attr (*__get_function_unwrapped(__attr context, __attr target))();
__attr (*__get_function_member(__attr target))();
__attr (*__check_and_get_function(__attr context, __attr target))();
__attr (*__check_and_get_function_unwrapped(__attr context, __attr target))();

/* Parameter position operations. */

int __HASPARAM(const __ptable *ptable, int ppos, int pcode);

/* Conversions. */

__attr __CONTEXT_AS_VALUE(__attr attr);

/* Type testing. */

__ref __ISFUNC(__ref obj);

#define __ISNULL(__ATTR) (!(__ATTR).value)

/* Attribute codes and positions for type objects. */

unsigned int __TYPECODE(__ref obj);
unsigned int __TYPEPOS(__ref obj);

/* Memory allocation. */

void *__ALLOCATE(size_t nmemb, size_t size);
void *__ALLOCATEIM(size_t nmemb, size_t size);
void *__REALLOCATE(void *ptr, size_t size);

/* Copying of structures. */

__ref __COPY(__ref obj, int size);

/* Result target administration for potential value replacement. */

__attr __set_attr(volatile __attr *target, __attr attr);

#endif /* __OPS_H__ */
